#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include "texinfo_parser.h"

#define INITIAL_STACK_SIZE 16
#define INITIAL_TOKEN_BUFFER_SIZE 128

#undef DEBUG

#ifdef DEBUG
#define dprintf(x) printf x ;
#else
#define dprintf(x)
#endif

/* The lexical state is recorded in a function name */
typedef int (*lexer_state_t)(
    texinfo_parser_t self,
    int ch);    

struct texinfo_parser
{
    /* The parser's callback */
    texinfo_parser_cb_t callback;

    /* The argument for the callback */
    void *rock;

    /* The state stack */
    int *state_stack;

    /* The end of the state stack */
    int *state_end;

    /* The top of the state stack */
    int *state_top;

    /* The value stack */
    char **value_stack;

    /* The top of the value stack */
    char **value_top;

    /* The current lexical state */
    lexer_state_t state;

    /* The token buffer */
    char *token;

    /* The end of the token buffer */
    char *token_end;

    /* The next character in the token buffer */
    char *point;

    /* The file being parsed */
    char *filename;

    /* The current line number of the input file */
    int line;

    /* Preserve formatting for an @example tag */
    int preserve;
};



/* Evaluate a "@@" command */
static char *perform_at(texinfo_parser_t self, char *cmd, char *args)
{
    return strdup("@");
}

/* Evaluate a "@{" command */
static char *perform_left_brace(texinfo_parser_t self, char *cmd, char *args)
{
    return strdup("{");
}

/* Evaluate a "@}" command */
static char *perform_right_brace(texinfo_parser_t self, char *cmd, char *args)
{
    return strdup("}");
}

/* Evaluate an "@'{}" command */
static char *perform_acute(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	if ((result = (char *)malloc(strlen(args) + 8)) == NULL)
	{
	    return NULL;
	}

	sprintf(result, "&%sacute;", args);
	return result;
    }

    return strdup("&acute;");
}

/* Evaluate a "@`{}" command */
static char *perform_grave(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	if ((result = (char *)malloc(strlen(args) + 8)) == NULL)
	{
	    return NULL;
	}

	sprintf(result, "&%sgrave;", args);
	return result;
    }

    return strdup("&lsquo");
}

/* Evaluate an "@^{}" command */
static char *perform_circumflex(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	if ((result = (char *)malloc(strlen(args) + 7)) == NULL)
	{
	    return NULL;
	}

	sprintf(result, "&%scirc;", args);
	return result;
    }

    return strdup("^");
}

/* Evaluate an "@"{}" command */
static char *perform_umlaut(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	if ((result = (char *)malloc(strlen(args) + 6)) == NULL)
	{
	    return NULL;
	}

	sprintf(result, "&%suml;", args);
	return result;
    }

    return strdup("&uml;");
}


/* Evalute a "@~{}" command */
static char *perform_tilde(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	if ((result = (char *)malloc(strlen(args) + 8)) == NULL)
	{
	    return NULL;
	}

	sprintf(result, "&%stilde;", args);
	return result;
    }

    return strdup("~");
}

/* Evaluate a "@,{}" command */
static char *perform_cedilla(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	if ((result = (char *)malloc(strlen(args) + 8)) == NULL)
	{
	    return NULL;
	}

	sprintf(result, "&%scedil;", args);
	return result;
    }

    return strdup("&cedil;");
}

/* Evaluate a "@={}" command */
static char *perform_macron(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	if ((result = (char *)malloc(strlen(args) + 8)) == NULL)
	{
	    return NULL;
	}

	sprintf(result, "%s&macr;", args);
	return result;
    }

    return strdup("&macr;");
}

/* Evaluate the @copyright{} command */
static char *perform_copyright(texinfo_parser_t self, char *cmd, char *args)
{
    return strdup("&copy;");
}

/* Evaluate the "@dots{}" command */
static char *perform_dots(texinfo_parser_t self, char *cmd, char *args)
{
    return strdup("<SMALL>...</SMALL>");
}

/* Evaluate an "@emph{}" command */
static char *perform_emph(texinfo_parser_t self, char *cmd, char *args)
{
    char *result;

    if (args == NULL)
    {
	return strdup("");
    }

    if ((result = (char *)malloc(strlen(args) + 8)) == NULL)
    {
	return NULL;
    }

    sprintf(result, "<I>%s</I>", args);
    return result;
}

/* Evaluate a "@strong{}" command */
static char *perform_strong(texinfo_parser_t self, char *cmd, char *args)
{
    char *result;

    if (args == NULL)
    {
	return strdup("");
    }

    if ((result = (char *)malloc(strlen(args) + 8)) == NULL)
    {
	return NULL;
    }

    sprintf(result, "<B>%s</B>", args);
    return result;
}

/* Evaluate the "@bullet{}" command */
static char *perform_bullet(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	result = malloc(strlen(args) + 2);
	sprintf(result, "*%s", args);
	return result;
    }

    return strdup("*");
}

/* Evaluate the "@minus{}" command */
static char *perform_minus(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	result = malloc(strlen(args) + 2);
	sprintf(result, "-%s", args);
	return result;
    }

    return strdup("-");
}

/* Perform the "@code" command */
static char *perform_code(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	result = malloc(strlen(args) + 14);
	sprintf(result, "<CODE>%s</CODE>", args);
	return result;
    }

    return strdup("");
}

/* Perform the "@uref" command */
static char *perform_uref(texinfo_parser_t self, char *cmd, char *args)
{
    char *result;
    char *point;

    /* Sanity check */
    if (args == NULL)
    {
	return strdup("");
    }

    /* If there's only the URL then use it as both text and link */
    if ((point = strchr(args, ',')) == NULL)
    {
	result = malloc(strlen(args) * 2 + 16);
	sprintf(result, "<A HREF=\"%s\">%s</A>", args, args);
	return result;
    }

    /* Otherwise split the thing up */
    result = malloc(strlen(args) * 2 + 16);
    strcpy(result, "<A HREF=\"");
    strncat(result, args, point - args);
    strcat(result, "\">");

    /* Skip over whitespace */
    point++;
    while (isspace((int)*point))
    {
	point++;
    }

    /* Finish up */
    strcat(result, point);
    strcat(result, "</A>");
    return result;
}

/* Perform the "@var" command */
static char *perform_var(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	result = malloc(strlen(args) + 12);
	sprintf(result, "<VAR>%s</VAR>", args);
	return result;
    }

    return strdup("");
}

/* Evaluate the "@samp{}" command */
static char *perform_samp(texinfo_parser_t self, char *cmd, char *args)
{
    char *result;
    
    result = malloc(strlen(args) + 15);
    sprintf(result, "&lsquo;%s&rsquo;", args);
    return result;
}

/* Perform an inline command */
static char *perform_curly_cmd(texinfo_parser_t self, char *cmd, char *args)
{
    /* Use the first letter of the command to limit the number of strcmps we do */
    switch (cmd[0])
    {
	case '"':
	{
	    if (strcmp(cmd, "\"") == 0)
	    {
		return perform_umlaut(self, cmd, args);
	    }

	    break;
	}

	case '\'':
	{
	    if (strcmp(cmd, "'") == 0)
	    {
		return perform_acute(self, cmd, args);
	    }

	    break;
	}

	case ',':
	{
	    if (strcmp(cmd, ",") == 0)
	    {
		return perform_cedilla(self, cmd, args);
	    }

	    break;
	}

	case '=':
	{
	    if (strcmp(cmd, "=") == 0)
	    {
		return perform_macron(self, cmd, args);
	    }

	    break;
	}

	case '^':
	{
	    if (strcmp(cmd, "^") == 0)
	    {
		return perform_circumflex(self, cmd, args);
	    }

	    break;
	}

	case '`':
	{
	    if (strcmp(cmd, "`") == 0)
	    {
		return perform_grave(self, cmd, args);
	    }

	    break;
	}

	case '@':
	{
	    if (strcmp(cmd, "@") == 0)
	    {
		return perform_at(self, cmd, args);
	    }

	    break;
	}

	case '{':
	{
	    if (strcmp(cmd, "{") == 0)
	    {
		return perform_left_brace(self, cmd, args);
	    }

	    break;
	}

	case '}':
	{
	    if (strcmp(cmd, "}") == 0)
	    {
		return perform_right_brace(self, cmd, args);
	    }

	    break;
	}

	case 'b':
	{
	    if (strcmp(cmd, "bullet") == 0)
	    {
		return perform_bullet(self, cmd, args);
	    }

	    break;
	}

	case 'c':
	{
	    if (strcmp(cmd, "code") == 0)
	    {
		return perform_code(self, cmd, args);
	    }

	    if (strcmp(cmd, "copyright") == 0)
	    {
		return perform_copyright(self, cmd, args);
	    }

	    break;
	}

	case 'd':
	{
	    if (strcmp(cmd, "dots") == 0)
	    {
		return perform_dots(self, cmd, args);
	    }

	    break;
	}

	case 'e':
	{
	    if (strcmp(cmd, "emph") == 0)
	    {
		return perform_emph(self, cmd, args);
	    }

	    break;
	}

	case 'm':
	{
	    if (strcmp(cmd, "minus") == 0)
	    {
		return perform_minus(self, cmd, args);
	    }

	    break;
	}

	case 's':
	{
	    if (strcmp(cmd, "samp") == 0)
	    {
		return perform_samp(self, cmd, args);
	    }

	    if (strcmp(cmd, "strong") == 0)
	    {
		return perform_strong(self, cmd, args);
	    }

	    break;
	}

	case 'u':
	{
	    if (strcmp(cmd, "uref") == 0)
	    {
		return perform_uref(self, cmd, args);
	    }
	}

	case 'v':
	{
	    if (strcmp(cmd, "var") == 0)
	    {
		return perform_var(self, cmd, args);
	    }

	    break;
	}

	case '~':
	{
	    if (strcmp(cmd, "~") == 0)
	    {
		return perform_tilde(self, cmd, args);
	    }

	    break;
	}
    }

    /* Warn of unknown commands */
    if (args == NULL)
    {
	fprintf(stderr, "%s:%d: Unknown command \"%s\".\n", self -> filename, self -> line, cmd);
	return strdup("");
    }

    fprintf(stderr, "%s:%d: Unknown command \"%s{%s}\".\n",
	    self -> filename, self -> line,
	    cmd, args);
    return strdup(args);
}


/* Perform the "@c" command */
static char *perform_comment(texinfo_parser_t self, char *cmd, char *args)
{
    return strdup("");
}

/* Perform the "@item" command */
static char *perform_item(texinfo_parser_t self, char *cmd, char *args)
{
    if (args != NULL)
    {
	char *result;

	if ((result = (char *)malloc(strlen(args) + 5)) == NULL)
	{
	    return NULL;
	}

	sprintf(result, "<LI>%s", args);
	return result;
    }

    return strdup("<LI>");
}

/* Perform the "@itemize" command */
static char *perform_itemize(texinfo_parser_t self, char *cmd, char *args)
{
    return strdup("<UL>\n");
}

/* Perform the "@enumerate" command */
static char *perform_enumerate(texinfo_parser_t self, char *cmd, char *args)
{
    return strdup("<OL>\n");
}

/* Perform the "@example" command */
static char *perform_example(texinfo_parser_t self, char *cmd, char *args)
{
    self->preserve++;
    return strdup("<PRE>\n");
}

/* Perform the "@end" command */
static char *perform_end(texinfo_parser_t self, char *cmd, char *args)
{
    /* Must have an arg */
    if (args == NULL)
    {
	fprintf(stderr, "%s:%d: \"@%s\" needs something after it.\n",
		self -> filename, self -> line, cmd);
	return strdup("");
    }

    /* FIX THIS: this should match up the end to the beginning somehow */
    if (strcmp(args, "itemize") == 0)
    {
	return strdup("</UL>\n");
    }

    if (strcmp(args, "enumerate") == 0)
    {
	return strdup("</OL>\n");
    }

    if (strcmp(args, "example") == 0)
    {
	if (self -> preserve > 0)
	{
	    self -> preserve--;
	}

	return strdup("</PRE>\n");
    }

    fprintf(stderr, "%s:%d: \"@%s\" not yet supported.\n", self -> filename, self -> line, cmd);
    return strdup("");
}

/* Perform line-terminated commands */
static char *perform_line_cmd(texinfo_parser_t self, char *cmd, char *args)
{
    /* Use the first letter of the command to limit the number of strcmps we do */
    switch (cmd[0])
    {
	case 'c':
	{
	    if (strcmp(cmd, "c") == 0)
	    {
		return perform_comment(self, cmd, args);
	    }

	    break;
	}

	case 'e':
	{
	    if (strcmp(cmd, "end") == 0)
	    {
		return perform_end(self, cmd, args);
	    }

	    if (strcmp(cmd, "enumerate") == 0)
	    {
		return perform_enumerate(self, cmd, args);
	    }

	    if (strcmp(cmd, "example") == 0)
	    {
		return perform_example(self, cmd, args);
	    }

	    break;
	}

	case 'i':
	{
	    if (strcmp(cmd, "item") == 0)
	    {
		return perform_item(self, cmd, args);
	    }

	    if (strcmp(cmd, "itemize") == 0)
	    {
		return perform_itemize(self, cmd, args);
	    }

	    break;
	}
    }

    /* Warn of unknown commands */
    fprintf(stderr, "%s:%d: Unknown command \"%s\".\n", self -> filename, self -> line, cmd);
    return strdup(args == NULL ? "" : args);
}


/* Reduction function declarations */
typedef char *(*reduction_t)(texinfo_parser_t self);

static char *append_lf_and_item(texinfo_parser_t self);
static char *append_item(texinfo_parser_t self);
static char *item_from_ws(texinfo_parser_t self);
static char *item_from_word(texinfo_parser_t self);
static char *item_from_curly(texinfo_parser_t self);
static char *item_from_noarg_curly(texinfo_parser_t self);
static char *item_from_cmd(texinfo_parser_t self);
static char *extend_lf_list(texinfo_parser_t self);
static char *make_lf_list(texinfo_parser_t self);
static char *make_cmd(texinfo_parser_t self);
static char *make_noarg_cmd(texinfo_parser_t self);
static char *make_cmd_item_list(texinfo_parser_t self);
static char *identity(texinfo_parser_t self);
static char *identity2(texinfo_parser_t self);

#include "texinfo.h"

/* <item-list> ::= <item-list> <lf-list> <item> */
static char *append_lf_and_item(texinfo_parser_t self)
{
    char *result;
    int length1;
    int length2;
    int length3;

    dprintf(("<item-list> ::= <item-list> <lf-list> <item>\n"));

    /* Concatenate the strings */
    length1 = strlen(self -> value_top[0]);
    length2 = strlen(self -> value_top[1]);
    length3 = strlen(self -> value_top[2]);
    if ((result = (char *)malloc(length1 + length2 + length3 + 1)) == NULL)
    {
	return NULL;
    }

    /* Copy the strings into place */
    memcpy(result, self -> value_top[0], length1);
    memcpy(result + length1, self -> value_top[1], length2);
    memcpy(result + length1 + length2, self -> value_top[2], length3 + 1);

    /* Clean up */
    free(self -> value_top[0]);
    free(self -> value_top[1]);
    free(self -> value_top[2]);
    return result;
}

/* <item-list> ::= <item-list> <item> */
/* <cmd-item-list> ::= <cmd-item-list> <cmd-item> */
static char *append_item(texinfo_parser_t self)
{
    char *result;
    int length1;
    int length2;

    dprintf(("<[cmd-]item-list> ::= <[cmd-]item-list> <[cmd-]item>\n"));

    /* Concatenate the strings */
    length1 = strlen(self -> value_top[0]);
    length2 = strlen(self -> value_top[1]);
    if ((result = (char *)malloc(length1 + length2 + 1)) == NULL)
    {
	return NULL;
    }

    /* Copy the strings into place */
    memcpy(result, self -> value_top[0], length1);
    memcpy(result + length1, self -> value_top[1], length2 + 1);

    /* Clean up */
    free(self -> value_top[0]);
    free(self -> value_top[1]);
    return result;
}

/* <item> ::= WS */
/* <cmd-item> ::= WS */
static char *item_from_ws(texinfo_parser_t self)
{
    dprintf(("<[cmd-]item> ::= WS\n"));
    return self -> value_top[0];
}

/* <item> ::= WORD */
/* <cmd-item> ::= WORD */
static char *item_from_word(texinfo_parser_t self)
{
    dprintf(("<[cmd-]item> ::= WORD\n"));
    return self -> value_top[0];
}

/* <item> ::= AT_WORD LBRACE <item-list> RBRACE */
/* <cmd-item> ::= AT_WORD LBRACE <cmd-item-list> RBRACE */
static char *item_from_curly(texinfo_parser_t self)
{
    char *result;

    /* Perform the command */
    dprintf(("<[cmd-]item> ::= AT_WORD LBRACE <[cmd-]item-list> RBRACE\n"));
    if ((result = perform_curly_cmd(self, self -> value_top[0], self -> value_top[2])) == NULL)
    {
	return NULL;
    }

    /* Clean up */
    free(self -> value_top[0]);
    free(self -> value_top[2]);
    return result;
}

/* <item> ::= AT_WORD LBRACE RBRACE */
/* <cmd-item> ::= AT_WORD LBRACE RBRACE */
/* <cmd-item> ::= AT_WORD */
static char *item_from_noarg_curly(texinfo_parser_t self)
{
    char *result;

    /* Perform the command */
    dprintf(("<[cmd-]item> ::= AT_WORD [LBRACE RBRACE]\n"));
    if ((result = perform_curly_cmd(self, self -> value_top[0], NULL)) == NULL)
    {
	return NULL;
    }

    /* Clean up */
    free(self -> value_top[0]);
    return result;
}

/* <item> ::= <command> */
static char *item_from_cmd(texinfo_parser_t self)
{
    dprintf(("<item> ::= <command>\n"));
    return self -> value_top[0];
}

/* <lf-list> ::= <lf-list> LF */
static char *extend_lf_list(texinfo_parser_t self)
{
    dprintf(("<lf-list> ::= <lf-list> LF\n"));
    /* Multiple linefeeds become paragraphs */
    if (*(self -> value_top[0]) == ' ')
    {
	free(self -> value_top[0]);
	return strdup("\n<P>");
    }

    return self -> value_top[0];
}

/* <lf-list> ::= LF */
static char *make_lf_list(texinfo_parser_t self)
{
    dprintf(("<lf-list> ::= LF\n"));

    if (self -> preserve)
    {
	return strdup("\n");
    }

    return strdup(" ");
}


/* <command> ::= AT_WORD <cmd-item-list> LF */
static char *make_cmd(texinfo_parser_t self)
{
    char *result;

    /* Perform the command */
    dprintf(("<command> ::= AT_WORD <cmd-item-list> LF\n"));
    if ((result = perform_line_cmd(self, self -> value_top[0], self -> value_top[1])) == NULL)
    {
	return NULL;
    }

    /* Clean up */
    free(self -> value_top[0]);
    free(self -> value_top[1]);
    return result;
}

/* <command> ::= AT_WORD LF */
static char *make_noarg_cmd(texinfo_parser_t self)
{
    char *result;

    /* Perform the command */
    dprintf(("<command> ::= AT_WORD LF\n"));
    if ((result = perform_line_cmd(self, self -> value_top[0], NULL)) == NULL)
    {
	return NULL;
    }

    /* Clean up */
    free(self -> value_top[0]);
    return result;
}

/* <cmd-item-list> ::= <cmd-item> */
static char *make_cmd_item_list(texinfo_parser_t self)
{
    /* Truncate initial whitespace */
    dprintf(("<cmd-item-list> ::= <cmd-item>\n"));
    if (isspace((int)*(self -> value_top[0])))
    {
	*(self -> value_top[0]) = '\0';
    }

    return self -> value_top[0];
}

/* <text> ::= <item-list> */
/* <item-list> ::= <item> */
static char *identity(texinfo_parser_t self)
{
    dprintf(("<text> ::= <item-list> | <item-list> ::= <item>\n"));
    return self -> value_top[0];
}

/* <text> ::= <item-list> <lf-list> */
/* <item-list> ::= <lf-list> <item> */
static char *identity2(texinfo_parser_t self)
{
    dprintf(("<item-list> ::= <lf-list> <item>\n"));

    /* Concatenate the items */
    if (self -> value_top[0])
    {
	char *result;
	int length1;
	int length2;

	/* Allocate enough space for the result */
	length1 = strlen(self -> value_top[0]);
	length2 = strlen(self -> value_top[1]);
	if ((result = (char *)malloc(length1 + length2 + 1)) == NULL)
	{
	    return NULL;
	}

	/* Copy the strings into place */
	memcpy(result, self -> value_top[0], length1);
	memcpy(result + length1, self -> value_top[1], length2 + 1);

	/* Clean up */
	free(self -> value_top[0]);
	free(self -> value_top[1]);
	return result;
    }
    else
    {
	return self -> value_top[1];
    }
}


/* Increase the size of the stack */
static int grow_stack(texinfo_parser_t self)
{
    dprintf(("grow_stack(): not yet implemented\n"));
    abort();
}

/* Push a state and value onto the stack */
static int push(texinfo_parser_t self, int state, void *value)
{
    /* Make sure there's enough room on the stack */
    if (! (self -> state_top + 1 < self -> state_end))
    {
	if (grow_stack(self) < 0)
	{
	    return -1;
	}
    }

    /* The state stack is pre-increment */
    *(++self -> state_top) = state;

    /* The value stack is post-increment */
    *(self -> value_top)++ = value;

    return 1;
}

/* Moves the top of the stack back "count" positions */
static void pop(texinfo_parser_t self, int count)
{
    self -> state_top -= count;
    self -> value_top -= count;
}

/* Puts n elements back onto the stack */
static void unpop(texinfo_parser_t self, int count)
{
    self -> state_top += count;
    self -> value_top += count;
}

/* Answers the top of the state stack */
static int top(texinfo_parser_t self)
{
    return *(self -> state_top);
}


/* Accept another token and move the parser state along until we need
 * to shift, accept or encounter an error */
static int shift_reduce(texinfo_parser_t self, terminal_t terminal, void *value)
{
    int action;

    /* Reduce as many times as possible */
    while (IS_REDUCE(action = sr_table[top(self)][terminal]))
    {
	struct production *production;
	int reduction;
	void *result;

	/* Locate the production rule to apply */
	reduction = REDUCTION(action);
	production = productions + reduction;

	/* Point the stack at the beginning of the 
	 * components of the production rule */
	pop(self, production -> count);

	/* Reduce by calling the production's reduction function */
	if ((result = production -> reduction(self)) == NULL)
	{
	    /* Put the args back onto the stack */
	    unpop(self, production -> count);
	    return 0;
	}

	/* Push the result of the reduction back onto the stack */
	if (push(self, REDUCE_GOTO(top(self), production), result) < 0)
	{
	    return -1;
	}
    }

    /* See if we can shift */
    if (IS_SHIFT(action))
    {
	return push(self, SHIFT_GOTO(action), value);
    }

    /* See if we can accept */
    if (IS_ACCEPT(action))
    {
	if (self -> callback != NULL)
	{
	    self -> callback(self -> rock, self -> value_stack[0]);
	}

	free(self -> value_stack[0]);
	return 0;
    }

    /* Everything else is an error */
    fprintf(stderr, "parse error in %s:%d\n", self -> filename, self -> line);
    exit(1);
}



/* Accept the EOF token */
static int accept_eof(texinfo_parser_t self)
{
    dprintf(("accept: EOF\n"));
    return shift_reduce(self, TT_EOF, NULL);
}

/* Accept an LBRACE token */
static int accept_lbrace(texinfo_parser_t self)
{
    dprintf(("accept: LBRACE\n"));
    return shift_reduce(self, TT_LBRACE, NULL);
}

/* Accept an RBRACE token */
static int accept_rbrace(texinfo_parser_t self)
{
    dprintf(("accept: RBRACE\n"));
    return shift_reduce(self, TT_RBRACE, NULL);
}

/* Accept a LF token */
static int accept_lf(texinfo_parser_t self, char *space)
{
    dprintf(("accept: LF\n"));
    return shift_reduce(self, TT_LF, NULL);
}

/* Accept a WS token */
static int accept_ws(texinfo_parser_t self, char *space)
{
    dprintf(("accept: WS\n"));
    return shift_reduce(self, TT_WS, strdup(space));
}

/* Accept a WORD token */
static int accept_word(texinfo_parser_t self, char *word)
{
    dprintf(("accept: WORD\n"));
    return shift_reduce(self, TT_WORD, strdup(word));
}

/* Accept an AT token */
static int accept_at_word(texinfo_parser_t self, char *word)
{
    dprintf(("accept: AT-WORD\n"));
    return shift_reduce(self, TT_AT_WORD, strdup(word));
}


/* Function prototypes for the lexer states */
static int lex_start(texinfo_parser_t self, int ch);
static int lex_ws(texinfo_parser_t self, int ch);
static int lex_word(texinfo_parser_t self, int ch);
static int lex_at_first(texinfo_parser_t self, int ch);
static int lex_at(texinfo_parser_t self, int ch);
static int lex_accented(texinfo_parser_t self, int ch);


/* Expands the token buffer */
static int grow_buffer(texinfo_parser_t self)
{
    size_t length = (self -> token_end - self -> token) * 2;
    char *token;

    /* Allocate a bigger buffer */
    if ((token = (char *)realloc(self -> token, length)) == NULL)
    {
	return -1;
    }

    /* Update the pointers */
    self -> point = token + (self -> point - self -> token);
    self -> token = token;
    self -> token_end = token + length;
    return 0;
}

/* Appends a character to the end of the token buffer */
static int append_char(texinfo_parser_t self, int ch)
{
    if (! (self -> point < self -> token_end))
    {
	if (grow_buffer(self) < 0)
	{
	    return -1;
	}
    }

    *(self -> point++) = ch;
    return 0;
}

/* Appends a string to the end of the token buffer */
static int append_string(texinfo_parser_t self, char *string)
{
    int ch;

    while ((ch = *(string++)) != 0)
    {
	if (append_char(self, ch) < 0)
	{
	    return -1;
	}
    }

    return 0;
}


/* Reading the first character of a token */
static int lex_start(texinfo_parser_t self, int ch)
{
    switch (ch)
    {
	case EOF:
	{
	    if (accept_eof(self) < 0)
	    {
		return -1;
	    }

	    self -> state = lex_start;
	    return 0;
	}

	case '@':
	{
	    self -> point = self -> token;
	    self -> state = lex_at_first;
	    return 0;
	}
 
	case '{':
	{
	    if (accept_lbrace(self) < 0)
	    {
		return -1;
	    }

	    self -> state = lex_start;
	    return 0;
	}

	case '}':
	{
	    if (accept_rbrace(self) < 0)
	    {
		return -1;
	    }

	    self -> state = lex_start;
	    return 0;
	}
    }

    /* Watch for whitespace */
    if (isspace(ch))
    {
	self -> point = self -> token;
	return lex_ws(self, ch);
    }

    /* Anything else is part of a word */
    self -> point = self -> token;
    return lex_word(self, ch);
}

/* Reading whitespace characters */
static int lex_ws(texinfo_parser_t self, int ch)
{
    /* A linefeed makes this a LF token */
    if (ch == '\n')
    {
	/* Append the linefeed to the token */
	if (append_char(self, ch) < 0)
	{
	    return -1;
	}

	/* Null-terminate the whitespace */
	if (append_char(self, 0) < 0)
	{
	    return -1;
	}

	/* Run it through the parser */
	if (accept_lf(self, self -> token) < 0)
	{
	    return -1;
	}

	self -> line++;
	self -> state = lex_start;
	return 0;
    }

    /* Other whitespace is just recorded */
    if (isspace(ch))
    {
	/* Record the character */
	if (append_char(self, ch) < 0)
	{
	    return -1;
	}

	self -> state = lex_ws;
	return 0;
    }

    /* Anything else is the end of the whitespace */
    if (append_char(self, 0) < 0)
    {
	return -1;
    }

    /* Run the whitespace through the parser */
    if (accept_ws(self, self -> token) < 0)
    {
	return -1;
    }

    /* Run the character through the start state */
    return lex_start(self, ch);
}

/* Reading letters of a word */
static int lex_word(texinfo_parser_t self, int ch)
{
    /* Watch for the end of the word */
    if (ch == '@' || ch == '{' || ch == '}' || isspace(ch))
    {
	/* Null-terminate the word */
	if (append_char(self, 0) < 0)
	{
	    return -1;
	}

	/* Run the word through the parser */
	if (accept_word(self, self -> token) < 0)
	{
	    return -1;
	}

	/* Run the next character through the start state */
	return lex_start(self, ch);
    }

    /* Transform those annoying HTML characters */
    switch (ch)
    {
	case '<':
	{
	    if (append_string(self, "&lt;") < 0)
	    {
		return -1;
	    }

	    break;
	}

	case '>':
	{
	    if (append_string(self, "&gt;") < 0)
	    {
		return -1;
	    }

	    break;
	}

	case '&':
	{
	    if (append_string(self, "&amp;") < 0)
	    {
		return -1;
	    }

	    break;
	}

	default:
	{
	    if (append_char(self, ch) < 0)
	    {
		return -1;
	    }

	    break;
	}
    }

    /* Stay in the same state */
    self -> state = lex_word;
    return 0;
}

/* Reading the first character after an "@" */
static int lex_at_first(texinfo_parser_t self, int ch)
{
    /* Watch for EOF */
    if (ch == EOF)
    {
	return -1;
    }

    /* Watch for special in-line @-words that don't require curly braces */
    switch (ch)
    {
	/* Some @-words are simply a single character */
	case '\t':
	case '\n':
	case ' ':
	case '*':
	case '.':
	case ':':
	case '@':
	case '{':
	case '}':
	{
	    /* Record the character for future reference */
	    if (append_char(self, ch) < 0)
	    {
		return -1;
	    }

	    /* Null-terminate it */
	    if (append_char(self, 0) < 0)
	    {
		return -1;
	    }

	    /* Accept it as an @-word */
	    if (accept_at_word(self, self -> token) < 0)
	    {
		return -1;
	    }

	    /* Insert curly braces to make the parsing easier */
	    if (accept_lbrace(self) < 0)
	    {
		return -1;
	    }

	    if (accept_rbrace(self) < 0)
	    {
		return -1;
	    }

	    self -> state = lex_start;
	    return 0;
	}

	/* Accented character prefixes can be non-alphabetic */
	case '"':
	case '\'':
	case ',':
	case '=':
	case '^':
	case '~':
	case '`':
	{
	    /* Record the accent for future reference */
	    if (append_char(self, ch) < 0)
	    {
		return -1;
	    }

	    /* Null-terminate it */
	    if (append_char(self, 0) < 0)
	    {
		return -1;
	    }

	    /* Accept it as an @-word */
	    if (accept_at_word(self, self -> token) < 0)
	    {
		return -1;
	    }

	    /* Read the accented character */
	    self -> state = lex_accented;
	    return 0;
	}
    }

    /* Anything else is the start of a normal @-word */
    if (append_char(self, ch) < 0)
    {
	return -1;
    }

    self -> state = lex_at;
    return 0;
}

/* Reading more characters of an @-word */
static int lex_at(texinfo_parser_t self, int ch)
{
    /* Watch for the end of the word */
    if (ch == EOF || ch == '@' || ch == '{' || ch == '}' || isspace(ch))
    {
	/* Null-terminate the @-word */
	if (append_char(self, 0) < 0)
	{
	    return -1;
	}

	/* Accept the @-word */
	if (accept_at_word(self, self -> token) < 0)
	{
	    return -1;
	}

	/* Run the next character through the start state */
	return lex_start(self, ch);
    }

    /* Anything else is part of the word */
    if (append_char(self, ch) < 0)
    {
	return -1;
    }

    self -> state = lex_at;
    return 0;
}

/* Read an accented character */
static int lex_accented(texinfo_parser_t self, int ch)
{
    /* Watch for curly braces around the accented letter */
    if (ch == '{')
    {
	/* Send the curly brace through the start state */
	return lex_start(self, ch);
    }

    /* Otherwise we need to insert the implied braces */
    if (accept_lbrace(self) < 0)
    {
	return -1;
    }

    /* If the character is not a alphabetical then leave the accent by itself */
    if (ch == EOF || ! isalpha(ch))
    {
	/* Insert the close brace */
	if (accept_rbrace(self) < 0)
	{
	    return -1;
	}

	/* Run ch through the start state */
	return lex_start(self, ch);
    }

    /* Otherwise record the character for later accenting */
    self -> point = self -> token;
    if (append_char(self, ch) < 0)
    {
	return -1;
    }

    /* Null-terminate it */
    if (append_char(self, 0) < 0)
    {
	return -1;
    }

    /* Accept it as a word */
    if (accept_word(self, self -> token) < 0)
    {
	return -1;
    }

    /* Insert the close brace */
    if (accept_rbrace(self) < 0)
    {
	return -1;
    }

    /* Go back to the start state */
    self -> state = lex_start;
    return 0;
}


/* Allocates and initializes a texinfo parser */
texinfo_parser_t texinfo_parser_alloc()
{
    texinfo_parser_t self;

    /* Allocate space for the parser */
    if ((self = (texinfo_parser_t)malloc(sizeof(struct texinfo_parser))) == NULL)
    {
	return NULL;
    }
    memset(self, 0, sizeof(struct texinfo_parser));

    /* Allocate room for the state stack */
    if ((self -> state_stack = (int *)calloc(INITIAL_STACK_SIZE, sizeof(int))) == NULL)
    {
	free(self);
	return NULL;
    }

    /* Allocate room for the value stack */
    if ((self -> value_stack = (char **)calloc(INITIAL_STACK_SIZE, sizeof(char *))) == NULL)
    {
	free(self -> state_stack);
	free(self);
	return NULL;
    }

    /* Allocate room for the initial token buffer */
    if ((self -> token = (char *)malloc(INITIAL_TOKEN_BUFFER_SIZE)) == NULL)
    {
	free(self -> value_stack);
	free(self -> state_stack);
	free(self);
	return NULL;
    }

    /* Initialize everything else */
    self -> state_end = self -> state_stack + INITIAL_STACK_SIZE;
    self -> token_end = self -> token + INITIAL_TOKEN_BUFFER_SIZE;
    texinfo_parser_reset(self, NULL, NULL, NULL);
    return self;
}

/* Releases the resources consumed by the receiver */
void texinfo_parser_free(texinfo_parser_t self)
{
    if (self -> filename != NULL)
    {
	free(self -> filename);
    }

    free(self -> token);
    free(self);
}

/* Resets the state of the texinfo parser */
void texinfo_parser_reset(texinfo_parser_t self, char *filename,
			  texinfo_parser_cb_t callback, void *rock)
{
    /* Dump the old filename */
    if (self -> filename != NULL)
    {
	free(self -> filename);
    }

    /* Record the new filename */
    if (filename != NULL)
    {
	self -> filename = strdup(filename);
    }
    else
    {
	self -> filename = NULL;
    }

    /* Reset everything else */
    self -> callback = callback;
    self -> rock = rock;
    self -> state_top = self -> state_stack;
    self -> value_top = self -> value_stack;
    self -> state = lex_start;
    self -> point = self -> token;
    self -> line = 1;
    self -> preserve = 0;

    /* Push a 0 state onto the top of the state stack */
    *(self -> state_top) = 0;
}

/* Parse some input */
int texinfo_parser_parse(texinfo_parser_t self, char *buffer, ssize_t length)
{
    char *pointer;

    /* Watch for the end of input marker */
    if (length == 0)
    {
	return self -> state(self, EOF);
    }

    /* Run everything through the lexer */
    for (pointer = buffer; pointer < buffer + length; pointer++)
    {
	if (self -> state(self, *pointer) < 0)
	{
	    return -1;
	}
    }

    return 0;
}
